"""
Cisco Terminal-Modes Module
"""

from aionet.vendors.terminal_modes.base import BaseTerminalMode


class EnableMode(BaseTerminalMode):
    """ Cisco Like Enable Mode Class """
    name = 'enable_mode'

    async def enter(self):
        """ Enter Enable Mode """
        self._logger.info("Entering to %s" % self.name)
        if await self.check():
            return ""
        output = await self.device.send_command(self._enter_command, pattern="Password")
        if "Password" in output:
            await self.device.send_command(self.device.secret)
        if not await self.check():
            raise ValueError("Failed to enter to %s" % self.name)
        self.device.current_terminal = self
        return output


class ConfigMode(BaseTerminalMode):
    """ Cisco Like Config Mode """
    name = 'config_mode'
    pass


class IOSxrConfigMode(ConfigMode):
    """ Cisco IOSxr Config Mode """

    async def exit(self):
        """Exit from configuration mode"""
        self._logger.info("Exiting from configuration mode")

        if not await self.check():
            return ""
        output = await self.device.send_command(self._exit_command,
                                                pattern=r"Uncommitted changes found")
        if "Uncommitted changes found" in output:
            output += await self.device.send_command("no")

        if await self.check(force=True):
            raise ValueError("Failed to exit from configuration mode")
        self.device.current_terminal = self._parent
        return output



"""Subclass specific to Aruba AOS 6.x"""

import re

from aionet.vendors.devices.base_ios import BaseIOSDevice


class ArubaAOS6(BaseIOSDevice):
    """Class for working with Aruba OS 6.X"""

    _disable_paging_command = "no paging"
    """Command for disabling paging"""

    _config_exit = "end"
    """Command for existing from configuration mode to privilege exec"""

    _config_check = ") (config"
    """Checking string in prompt. If it's exist im prompt - we are in configuration mode"""

    _pattern = r"\({prompt}.*?\) (\(.*?\))?\s?[{delimiters}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    async def _set_base_prompt(self):
        """
        Setting two important vars:

            base_prompt - textual prompt in CLI (usually hostname)
            base_pattern - regexp for finding the end of command. It's platform specific parameter

        For Aruba AOS 6 devices base_pattern is "(prompt) (\(.*?\))?\s?[#|>]
        """
        self._logger.info("Setting base prompt")
        prompt = await self._find_prompt()

        # Strip off trailing terminator
        self._base_prompt = prompt[1:-3]
        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        base_prompt = re.escape(self._base_prompt[:12])
        pattern = type(self)._pattern
        self._base_pattern = pattern.format(prompt=base_prompt, delimiters=delimiters)
        self._logger.debug("Base Prompt: %s" % self._base_prompt)
        self._logger.debug("Base Pattern: %s" % self._base_pattern)
        return self._base_prompt



"""Subclass specific to Fujitsu Blade Switch"""

import re

from aionet.vendors.devices.base_ios import BaseIOSDevice


class FujitsuSwitch(BaseIOSDevice):
    """Class for working with Fujitsu Blade switch"""

    _pattern = r"\({prompt}.*?\) (\(.*?\))?[{delimiters}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    _disable_paging_command = "no pager"
    """Command for disabling paging"""

    _config_enter = "conf"
    """Command for entering to configuration mode"""

    async def _set_base_prompt(self):
        """
        Setting two important vars
            base_prompt - textual prompt in CLI (usually hostname)
            base_pattern - regexp for finding the end of command. IT's platform specific parameter

        For Fujitsu devices base_pattern is "(prompt) (\(.*?\))?[>|#]"
        """
        self._logger.info("Setting base prompt")
        prompt = await self._find_prompt()
        # Strip off trailing terminator
        self._base_prompt = prompt[1:-3]
        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        base_prompt = re.escape(self._base_prompt[:12])
        pattern = type(self)._pattern
        self._base_pattern = pattern.format(prompt=base_prompt, delimiters=delimiters)
        self._logger.debug("Base Prompt: %s" % self._base_prompt)
        self._logger.debug("Base Pattern: %s" % self._base_pattern)
        return self._base_prompt

    @staticmethod
    def _normalize_linefeeds(a_string):
        """
        Convert '\r\r\n','\r\n', '\n\r' to '\n and remove extra '\n\n' in the text
        """
        newline = re.compile(r"(\r\r\n|\r\n|\n\r)")
        return newline.sub("\n", a_string).replace("\n\n", "\n")



"""
Terminal Modes Classes, which handle entering and exist to
different terminal modes
"""


class BaseTerminalMode:
    """ Base Terminal Mode """
    name = ''

    def __init__(self,
                 enter_command,
                 exit_command,
                 check_string,
                 device,
                 parent=None):
        """

        :param enter_command: Command to enter to the terminal mode (ex: conf t)
        :param exit_command: Command to exist the terminal mode (ex: end)
        :param check_string: string to check if the device in this terminal mode
        :param device: Device Object
        :param parent: parent Terminal, for example the enable mode is parent of config mode
        :type BaseTerminalMode
        """
        self._enter_command = enter_command
        self._exit_command = exit_command
        self._check_string = check_string
        self.device = device
        self._parent = parent

    def __eq__(self, other):
        """ Compare different terminal objects """
        if isinstance(other, type(self)):
            if self.name == other.name:
                return True
        return False

    async def __call__(self):
        """ callable terminal to enter """
        return await self.enter()

    @property
    def _logger(self):
        return self.device._logger

    async def check(self, force=False):
        """Check if are in configuration mode. Return boolean"""
        if self.device.current_terminal is not None and not force:
            if self.device.current_terminal == self:
                return True
        output = await self.device.send_new_line()
        return self._check_string in output

    async def enter(self):
        """ enter terminal mode """
        self._logger.info("Entering to %s" % self.name)
        if await self.check():
            return ""
        output = await self.device.send_command_expect(self._enter_command, pattern="Password")
        if not await self.check(force=True):
            raise ValueError("Failed to enter to %s" % self.name)
        self.device.current_terminal = self
        return output

    async def exit(self):
        """ exit terminal mode """
        self._logger.info("Exiting from %s" % self.name)
        if not await self.check():
            return ""
        if self.device.current_terminal != self:
            return ""

        output = await self.device.send_command_expect(self._exit_command)
        if await self.check(force=True):
            raise ValueError("Failed to Exit from %s" % self.name)
        self.device.current_terminal = self._parent
        return output

    async def send_command(self,
                           command_string,
                           pattern="",
                           re_flags=0,
                           strip_command=True,
                           strip_prompt=True,
                           ):
        """ API to send commands on this terminal """
        await self.enter()

        output = await self.device.send_command(
            command_string,
            pattern,
            re_flags,
            strip_command,
            strip_prompt,
        )
        return output



class BaseAionetError(Exception):
    _error_name = ''

    def __init__(self, ip_address, code, reason):
        self.ip_address = ip_address
        self.code = code
        self.reason = reason
        self.msg = "Host %s %s Error: %s" % (ip_address, type(self)._error_name, reason)
        super().__init__(self.msg)


class AionetAuthenticationError(BaseAionetError):
    _error_name = 'Authentication'


class AionetTimeoutError(BaseAionetError):
    _error_name = 'timeout'


class AionetCommitError(BaseAionetError):
    _error_name = 'commit'


class AionetConnectionError(BaseAionetError):
    _error_name = 'connection'



import re

from aionet.vendors.devices.base import BaseDevice


class Terminal(BaseDevice):
    """Class for working with General Terminal"""

    def __init__(self, delimeter_list=None, *args, **kwargs):
        """
        Initialize class for asynchronous working with network devices
        Invoke init with some special params (base_pattern and username)

        :param str host: device hostname or ip address for connection
        :param str username: username for logging to device
        :param str password: user password for logging to device
        :param int port: ssh port for connection. Default is 22
        :param str device_type: network device type
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param delimeter_list: list with delimeters
        :param str local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param str passphrase: password for encrypted client keys
        :param float timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object
        """
        super().__init__(*args, **kwargs)
        if delimeter_list is not None:
            self._delimiter_list = delimeter_list

    _delimiter_list = ["$", "#"]
    """All this characters will stop reading from buffer. It mean the end of device prompt"""

    _pattern = r"[{delimiters}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    async def _set_base_prompt(self):
        """Setting base pattern"""
        self._logger.info("Setting base prompt")
        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        pattern = type(self)._pattern
        base_pattern = pattern.format(delimiters=delimiters)
        self._logger.debug("Base Pattern: %s"% base_pattern)
        self._conn.set_base_pattern(base_pattern)



"""
Juniper Terminal Modes
"""
from aionet.vendors.terminal_modes.base import BaseTerminalMode
from aionet.vendors.terminal_modes.cisco import ConfigMode as CiscoConfigMode


class ConfigMode(CiscoConfigMode):
    async def check(self, force=False):
        """Check if are in configuration mode. Return boolean"""
        if self.device.current_terminal is not None and not force:
            if self.device.current_terminal == self:
                return True
        await self.device.send_new_line()
        output = await self.device.send_new_line()

        return self._check_string in output
    pass


class CliMode(BaseTerminalMode):
    name = 'cli_mode'

    def exit(self):
        pass



from aionet.dispatcher import ConnectionHandler, platforms
from aionet.exceptions import AionetAuthenticationError, AionetTimeoutError, AionetCommitError
from aionet.logging import logger
from aionet.version import __author__, __author_email__, __url__, __version__

__all__ = (
    "ConnectionHandler",
    "platforms",
    "logger",
    "AionetAuthenticationError",
    "AionetTimeoutError",
    "AionetCommitError",
    "vendors",
)



"""
JunOSLikeDevice Class is abstract class for using in Juniper JunOS like devices

Connection Method are based upon AsyncSSH and should be running in asyncio loop
"""

import re

from aionet.vendors.terminal_modes.juniper import ConfigMode
from aionet.vendors.devices.base import BaseDevice


class BAseJunOSDevice(BaseDevice):
    """
    JunOSLikeDevice Class for working with Juniper JunOS like devices

    Juniper JunOS like devices having several concepts:

    * shell mode (csh). This is csh shell for FreeBSD. This mode is not covered by this Class.
    * cli mode (specific shell). The entire configuration is usual configured in this shell:

      * operation mode. This mode is using for getting information from device
      * configuration mode. This mode is using for configuration system
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.current_terminal = None  # State Machine for the current Terminal mode of the session
        self.config_mode = ConfigMode(
            enter_command=type(self)._config_enter,
            exit_command=type(self)._config_check,
            check_string=type(self)._config_exit,
            device=self
        )

    _delimiter_list = ["%", ">", "#"]
    """All this characters will stop reading from buffer. It mean the end of device prompt"""

    _pattern = r"\w+(\@[\-\w]*)?[{delimiters}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    _disable_paging_command = "set cli screen-length 0"
    """Command for disabling paging"""

    _config_enter = "configure"
    """Command for entering to configuration mode"""

    _config_exit = "exit configuration-mode"
    """Command for existing from configuration mode to privilege exec"""

    _config_check = "#"
    """Checking string in prompt. If it's exist im prompt - we are in configuration mode"""

    _commit_command = "commit"
    """Command for committing changes"""

    _commit_comment_command = "commit comment {}"
    """Command for committing changes with comment"""

    async def _set_base_prompt(self):
        """
        Setting two important vars
            base_prompt - textual prompt in CLI (usually username or hostname)
            base_pattern - regexp for finding the end of command. IT's platform specific parameter

        For JunOS devices base_pattern is "user(@[hostname])?[>|#]
        """
        self._logger.info("Setting base prompt")
        prompt = await self._find_prompt()
        prompt = prompt[:-1]
        # Strip off trailing terminator
        if "@" in prompt:
            prompt = prompt.split("@")[1]
        self.device_prompt = prompt
        delimiters = map(re.escape, type(self)._delimiter_list)
        delimiters = r"|".join(delimiters)
        base_prompt = re.escape(self.device_prompt[:12])
        pattern = type(self)._pattern
        base_pattern = pattern.format(delimiters=delimiters)
        self._logger.debug("Base Prompt: %s" % self.device_prompt)
        self._logger.debug("Base Pattern: %s" % base_pattern)
        self.prompt_pattern = base_pattern
        self._conn.set_base_pattern(base_pattern)
        return self.device_prompt

    async def send_config_set(
            self,
            config_commands=None,
            with_commit=True,
            commit_comment="",
            exit_config_mode=True,
    ):
        """
        Sending configuration commands to device
        By default automatically exits/enters configuration mode.

        :param list config_commands: iterable string list with commands for applying to network devices in system view
        :param bool with_commit: if true it commit all changes after applying all config_commands
        :param string commit_comment: message for configuration commit
        :param bool exit_config_mode: If true it will quit from configuration mode automatically
        :return: The output of these commands
        """

        if config_commands is None:
            return ""

        # Send config commands
        output = await self.config_mode()
        output += await super().send_config_set(config_commands=config_commands)
        if with_commit:
            commit = type(self)._commit_command
            if commit_comment:
                commit = type(self)._commit_comment_command.format(commit_comment)

            output += await self.send_command_expect(commit)

        if exit_config_mode:
            output += await self.config_mode.exit()

        self._logger.debug("Config commands output: %s" % repr(output))

        return output



import setuptools
from aionet import __version__, __author__

try:
    # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # for pip <= 9.0.3
    from pip.req import parse_requirements


def load_requirements(fname):
    reqs = parse_requirements(fname, session="test")
    return [str(ir.req) for ir in reqs]


with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aionet",
    version=__version__,
    author=__author__,
    author_email="aaqrabaw@gmail.com",
    description="async networking SDK",
    long_description=long_description,
    url="https://github.com/Ali-aqrabawi/aionet",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=load_requirements("requirements.txt"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Logging Module
"""
import logging


class aionetLoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger, extra):
        super(aionetLoggerAdapter, self).__init__(logger, extra)

    def process(self, msg, kwargs):
        msg, kwargs = super(aionetLoggerAdapter, self).process(msg, kwargs)
        msg = 'Host %s: ' % kwargs['extra']['host'] + msg
        return msg, {}


logger = logging.getLogger(__package__)
logger.setLevel(logging.WARNING)



"""
ComwareLikeDevice Class is abstract class for using in HP Comware like devices

Connection Method are based upon AsyncSSH and should be running in asyncio loop
"""

import re

from aionet.vendors.terminal_modes.hp import SystemView
from aionet.vendors.devices.base import BaseDevice


class BaseComwareDevice(BaseDevice):
    """
    This Class for working with hp comware like devices

    HP Comware like devices having several concepts:

    * user exec or user view. This mode is using for getting information from device
    * system view. This mode is using for configuration system
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.current_terminal = None  # State Machine for the current Terminal mode of the session
        self.system_view = SystemView(
            enter_command=type(self)._system_view_enter,
            exit_command=type(self)._system_view_exit,
            check_string=type(self)._system_view_check,
            device=self
        )

    _delimiter_list = [">", "]"]
    """All this characters will stop reading from buffer. It mean the end of device prompt"""

    _delimiter_left_list = ["<", "["]
    """Begging prompt characters. Prompt must contain it"""

    _pattern = r"[{delimiter_left}]{prompt}[\-\w]*[{delimiter_right}]"
    """Pattern for using in reading buffer. When it found processing ends"""

    _disable_paging_command = "screen-length disable"
    """Command for disabling paging"""

    _system_view_enter = "system-view"
    """Command for entering to system view"""

    _system_view_exit = "return"
    """Command for existing from system view to user view"""

    _system_view_check = "]"
    """Checking string in prompt. If it's exist im prompt - we are in system view"""

    async def _session_preparation(self):
        """ Prepare Session """
        await super()._session_preparation()
        await self._disable_paging()

    async def _set_base_prompt(self):
        """
        Setting two important vars
            base_prompt - textual prompt in CLI (usually hostname)
            base_pattern - regexp for finding the end of command. IT's platform specific parameter

        For Comware devices base_pattern is "[\]|>]prompt(\-\w+)?[\]|>]
        """
        self._logger.info("Setting base prompt")
        prompt = await self._find_prompt()
        # Strip off trailing terminator
        self._base_prompt = prompt[1:-1]
        delimiter_right = map(re.escape, type(self)._delimiter_list)
        delimiter_right = r"|".join(delimiter_right)
        delimiter_left = map(re.escape, type(self)._delimiter_left_list)
        delimiter_left = r"|".join(delimiter_left)
        base_prompt = re.escape(self._base_prompt[:12])
        pattern = type(self)._pattern
        self._base_pattern = pattern.format(
            delimiter_left=delimiter_left,
            prompt=base_prompt,
            delimiter_right=delimiter_right,
        )
        self._logger.debug("Base Prompt: %s" % self._base_prompt)
        self._logger.debug("Base Pattern: %s" % self._base_pattern)
        return self._base_prompt

    async def send_config_set(self, config_commands=None, exit_system_view=False):
        """
        Sending configuration commands to device
        Automatically exits/enters system-view.

        :param list config_commands: iterable string list with commands for applying to network devices in system view
        :param bool exit_system_view: If true it will quit from system view automatically
        :return: The output of this commands
        """

        if config_commands is None:
            return ""

        # Send config commands
        output = await self.system_view()
        output += await super().send_config_set(config_commands=config_commands)

        if exit_system_view:
            output += await self.system_view.exit()

        output = self._normalize_linefeeds(output)
        self._logger.debug("Config commands output: %s" % repr(output))
        return output



import re

from aionet.vendors.devices.base_ios import BaseIOSDevice


class CiscoNXOS(BaseIOSDevice):
    """Class for working with Cisco Nexus/NX-OS"""

    @staticmethod
    def _normalize_linefeeds(a_string):
        """
        Convert '\r\n' or '\r\r\n' to '\n, and remove extra '\r's in the text
        """
        newline = re.compile(r"(\r\r\n|\r\n)")

        return newline.sub("\n", a_string).replace("\r", "")



"""
IOSLikeDevice Class is abstract class for using in Cisco IOS like devices

Connection Method are based upon AsyncSSH and should be running in asyncio loop
"""

from aionet.vendors.devices.base import BaseDevice
from aionet.vendors.terminal_modes.cisco import EnableMode, ConfigMode


class BaseIOSDevice(BaseDevice):
    """
    This Class is abstract class for working with Cisco IOS like devices

    Cisco IOS like devices having several concepts:

    * user exec or unprivileged exec. This mode allows you perform basic tests and get system information.
    * privilege exec. This mode allows the use of all EXEC mode commands available on the system
    * configuration mode or config mode. This mode are used for configuration whole system.
    """

    def __init__(self, secret=u"", *args, **kwargs):
        """
        Initialize class for asynchronous working with network devices

        :param str host: device hostname or ip address for connection
        :param str username: username for logging to device
        :param str password: user password for logging to device
        :param str secret: secret password for privilege mode
        :param int port: port number. Default is 22 for ssh and 23 for telnet
        :param str device_type: network device type
        :param known_hosts: file with known hosts. Default is None (no policy). With () it will use default file
        :param str local_addr: local address for binding source of tcp connection
        :param client_keys: path for client keys. Default in None. With () it will use default file in OS
        :param str passphrase: password for encrypted client keys
        :param float timeout: timeout in second for getting information from channel
        :param loop: asyncio loop object
        """
        super().__init__(*args, **kwargs)
        self.secret = secret

        self.current_terminal = None  # State Machine for the current Terminal mode of the session

        self.enable_mode = EnableMode(
            enter_command=type(self)._priv_enter,
            exit_command=type(self)._priv_exit,
            check_string=type(self)._priv_check,
            device=self
        )
        self.config_mode = ConfigMode(
            enter_command=type(self)._config_enter,
            exit_command=type(self)._config_exit,
            check_string=type(self)._config_check,
            device=self,
            parent=self.enable_mode
        )

    _priv_enter = "enable"
    """Command for entering to privilege exec"""

    _priv_exit = "disable"
    """Command for existing from privilege exec to user exec"""

    _priv_check = "#"
    """Checking string in prompt. If it's exist im prompt - we are in privilege exec"""

    _config_enter = "conf t"
    """Command for entering to configuration mode"""

    _config_exit = "end"
    """Command for existing from configuration mode to privilege exec"""

    _config_check = ")#"
    """Checking string in prompt. If it's exist im prompt - we are in configuration mode"""

    _disable_paging_command = "terminal length 0"
    """Command for disabling paging"""

    _disable_width_command = "terminal width 511"
    """Command for disabling paging"""

    async def _session_preparation(self):
        await super()._session_preparation()
        await self.enable_mode()
        await self._disable_paging()
        await self._disable_width()

    async def _disable_paging(self):
        """ disable terminal pagination """
        self._logger.info(
            "Disabling Pagination, command = %r" % type(self)._disable_paging_command)
        await self.send_command_expect(type(self)._disable_paging_command)

    async def _disable_width(self):
        """ disable width pagination """
        self._logger.info(
            "Disabling width, command = %r" % type(self)._disable_width_command)
        await self.send_command_expect(type(self)._disable_width_command)

    async def send_config_set(self, config_commands=None, exit_config_mode=True):
        """
        Sending configuration commands to Cisco IOS like devices
        Automatically exits/enters configuration mode.

        :param list config_commands: iterable string list with commands for applying to network devices in conf mode
        :param bool exit_config_mode: If true it will quit from configuration mode automatically
        :return: The output of this commands
        """

        if config_commands is None:
            return ""

        # Send config commands
        await self.config_mode()
        output = ''
        output += await super().send_config_set(config_commands=config_commands)

        if exit_config_mode:
            output += await self.config_mode.exit()

        self._logger.debug("Config commands output: %s" % repr(output))

        return output

    async def _cleanup(self):
        """ Any needed cleanup before closing connection """
        self._logger.info("Cleanup session")
        await self.config_mode.exit()
